from iec_parser import parse_iec_array
from processor import Processor
from numpy import array
from threading import Thread
from event_logger import *
import logging
import settings

class Flow:

  def __init__(self, src_addr, dst_addr, src_port, dst_port):
    self.processor = Processor(src_addr, dst_addr, src_port, dst_port)
    self.data = []
    self.src_addr = src_addr
    self.dst_addr = dst_addr
    self.src_port = src_port
    self.dst_port = dst_port
    self.id = src_addr + ':' + str(src_port) + '-' + dst_addr + ':' + str(dst_port)
    self.processing = False

  def process_async(self, data):
    logging.debug(self.id + ' Processing started')
    anomaly_count = 0
    for i in range(0, settings.detection_reiteration):
      if self.processor.process(data):
        anomaly_count += 1
        logging.info('Anomaly in iteration ' + str(i))
      else:
        logging.info('Normal operation in iteration ' + str(i))
    if anomaly_count >= settings.detection_reiteration:
      on_anomaly(self.src_addr, self.dst_addr, self.src_port, self.dst_port)
    else:
      on_normal_operation(self.src_addr, self.dst_addr, self.src_port, self.dst_port)
    self.processing = False;
    logging.debug(self.id + ' Processing done')

  def push(self, data):
    logging.debug('Pushed ' + str(len(data)) + ' bytes from flow ' + self.id)
    values = parse_iec_array(data)
    if len(values) >= 2:
      self.data.append(values)
      if len(self.data) >= settings.queue_size:
        if self.processing:
          logging.info(self.id + ' Queue full while processing')
          del self.data[0]
        else:
          tmp_data = array(self.data)
          thread = Thread(target = self.process_async, args = [tmp_data])
          logging.info(self.id + ' Processing scheduled asyncronousely')
          self.processing = True;
          thread.start()
          self.data = []
