import settings
import numpy as np
#from numpy import array, append, asarray, argmax, savetxt, concatenate, shape
from sklearn import mixture
from sklearn.decomposition import PCA
import logging

class Processor:

    def __init__(self, src_addr, dst_addr, src_port, dst_port):
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.src_port = src_port
        self.dst_port = dst_port
        self.id = src_addr + ':' + str(src_port) + '-' + dst_addr + ':' + str(dst_port)
        self.threshold = settings.detection_threshold * settings.queue_size / 100.0
        np.random.seed(1)
        np.set_printoptions(precision=4)

    def _mahal(self, gmmodel, X):
        Sigma=gmmodel.covars_
        mu=gmmodel.means_
        rmu=np.shape(mu)[0]
        rx=np.shape(X)[0]
        cx=np.shape(X)[1]
        mahalD=np.zeros([rx,rmu])
        for j in range(0,rmu):
            L = np.linalg.cholesky(Sigma[j,:,:])
            Xcentered=np.zeros([rx,cx])
            for rk in range(0,rx):
                Xcentered[rk,:]=X[rk,:]-mu[j,:]
            A=np.matrix(Xcentered)
            B=np.matrix(L)
            xRinv= A*((B.T).I)
            xRinvSq=np.square(xRinv)
            mahalD[:,j] = np.sum(np.array(xRinvSq),1)
        return mahalD

    def _normalization(self, data):
        d = data
        for k in range(0,2):
            d[:,k] = (data[:,k]-min(data[:,k]))/(max(data[:,k])-min(data[:,k]))
            data[:,k] = d[:,k]
        return data

    def clust_mahal(self, data):
        # always the first bus is considered as the reference bus
        # ref bus data [[phase, voltage],[phase, voltage]]
        ref_bus_data = np.concatenate((np.array([data[:, 1]]).T, np.array([data[:, 0]]).T),axis=1)
        cols = np.shape(data)[1]
        rows = np.shape(data)[0]

        #normalization
        ref_bus_data = self._normalization(ref_bus_data)

        #gmm clustering
        g = mixture.GMM(n_components=1, covariance_type='full', verbose=0)
        g.fit(ref_bus_data)
        #print g.weights_, g.means_, g.covars_

        #mahalanobis distance
        dist = self._mahal(g, ref_bus_data)
        ref_dist = np.max([np.max(dist), 3 * np.mean(dist)]);
        # print ref_dist

        bus = 0
        rez = np.zeros([cols / 2])
        #distM = np.zeros([cols / 2])
        for ind in range(0, cols, 2):
            tmp_bus_data = np.concatenate((np.array([data[:, ind + 1]]).T, np.array([data[:, ind]]).T),axis=1)
            tmp_bus_data = self._normalization(tmp_bus_data)
            tmp_dist = self._mahal(g, tmp_bus_data)
            alerts = tmp_dist > ref_dist;
            rez[bus] = np.sum(alerts);
            #distM[bus] = np.max(tmp_dist);
            bus = bus + 1

        return np.max(rez) > self.threshold

    def _pca_analysis(self, X):
        #pca analysis
        pca = PCA()
        pca.fit(X)
        coeff = pca.components_
        scores = pca.transform(X)
        return (coeff, scores)

    def _compute_pca_scores(self, X):
        coeff,scores = self._pca_analysis(X)
        wref = np.sum(coeff, 1);
        for k in range(0,2):
            scores[:,k] = scores[:,k] / wref[k]
        return scores

    def clust_mahal_pca(self, data):
        # always the first bus is considered as the reference bus
        # ref bus data [[phase, voltage],[phase, voltage]]
        ref_bus_data = np.concatenate((np.array([data[:, 1]]).T, np.array([data[:, 0]]).T),axis=1)
        cols = np.shape(data)[1]
        rows = np.shape(data)[0]

        #normalization
        ref_bus_data = self._normalization(ref_bus_data)

        #pca analysis
        scores = self._compute_pca_scores(ref_bus_data)

        #gmm clustering
        g = mixture.GMM(n_components=1, covariance_type='full', verbose=0)
        g.fit(scores)
        #print g.weights_, g.means_, g.covars_

        #mahalanobis distance
        dist = self._mahal(g, scores)
        ref_dist = np.max([np.max(dist), 3 * np.mean(dist)]);
        # print ref_dist

        all_features = np.zeros([rows, cols])
        for ind in range(0, cols, 2):
            tmp_bus_data = np.concatenate((np.array([data[:, ind + 1]]).T, np.array([data[:, ind]]).T),axis=1)
            tmp_bus_data = self._normalization(tmp_bus_data)
            all_features[:, ind:ind + 2] = tmp_bus_data

        scores = self._compute_pca_scores(all_features)
        scores = scores[:, 0:2]

        tmp_dist = self._mahal(g, scores)
        alerts = tmp_dist > ref_dist;
        print alerts

        rez = np.sum(alerts);

        print rez
        return np.max(rez) > self.threshold

    def process(self, data_matrix):
        logging.debug('Processing\n' + str(data_matrix) + '\nfor ' + self.id)
        #return self.clust_mahal(data_matrix)
        return self.clust_mahal_pca(data_matrix)